#!/usr/local/bin/perl
# 
# Object class for manipulating AKAI S3000 disk image.
# Copyright (c) 1997, Hiroyuki Ohsaki.
# All rights reserved.
# 
# $Id: AkaiDisk.pm,v 1.8 1998/10/29 06:40:41 oosaki Exp ohsaki $
# 

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

package Synth::AkaiDisk;

require 5.000;
use File::Basename;
use Symbol;
use Carp;
use POSIX;
use strict;

use vars qw(@ISA @EXPORT @EXPORT_OK
	    $BLOCK_SIZE
	    $MAX_FILE_ENTRIES_S1000
	    $MAX_FILE_ENTRIES_S3000
	    $FILE_ENTRY_SIZE
	    $MAX_DIR_ENTRIES
	    $DIR_ENTRY_OFFSET
	    $DIR_ENTRY_SIZE
	    $FAT_OFFSET
	    $FAT_MARK_RESERVED
	    $FAT_MARK_RESERVED2
	    $FAT_MARK_FREE
	    $FAT_MARK_END
	    $ROOT_ENTRY_OFFSET
	    $TYPE_FREE
	    $TYPE_DIR_S1000
	    $TYPE_DIR_S3000
	    $MAX_PART_SIZE
	    $DISK_HEADER
	    $PART_HEADER
	    $PART_HEADER_OFFSET
	    $PART_TABLE_OFFSET
	    $PART_END_MARK
	    $VERSION
	    $DEBUG);

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(qualify_path unqualify_path join_path);

$BLOCK_SIZE = 0x2000;
$MAX_FILE_ENTRIES_S1000 = 125;	# should be 128 ??
$MAX_FILE_ENTRIES_S3000 = 509;	# should be 512 ??
$FILE_ENTRY_SIZE = 24;
$MAX_DIR_ENTRIES = 100;
$DIR_ENTRY_OFFSET = 0xca;	# start address of the directory entry
$DIR_ENTRY_SIZE = 16;
$FAT_OFFSET = 0x70a;		# start address of the file allocation table
$FAT_MARK_RESERVED = 0x4000;	# reserved by the system
$FAT_MARK_RESERVED2 = 0x8000;	# reserved for second directory entry
$FAT_MARK_FREE = 0;
$FAT_MARK_END = 0xc000;		# end code of a FAT chain
$ROOT_ENTRY_OFFSET = 0;		# start address of root directory
$TYPE_FREE = 0;
$TYPE_DIR_S1000 = 1;
$TYPE_DIR_S3000 = 3;
$MAX_PART_SIZE = 60;		# maximum parition size (in Mbyte)
$DISK_HEADER = <<EOF;
00 00 05 0d 0a 1a 0f 27 14 34 19 41 1e 4e
23 5b 28 68 2d 75 32 82 37 8f 3c 9c 41 a9 46 b6
4b c3 50 d0 55 dd 5a ea 5f f7 64 04 69 11 6e 1e
73 2b 78 38 7d 45 82 52 87 5f 8c 6c 91 79 96 86
9b 93 a0 a0 a5 ad aa ba af c7 b4 d4 b9 e1 be ee
c3 fb c8 08 cd 15 d2 22 d7 2f dc 3c e1 49 e6 56
eb 63 f0 70 f5 7d fa 8a ff 97 04 a5 09 b2 0e bf
13 cc 18 d9 1d e6 22 f3 27 00 2c 0d 31 1a 36 27
3b 34 40 41 45 4e 4a 5b 4f 68 54 75 59 82 5e 8f
63 9c 68 a9 6d b6 72 c3 77 d0 7c dd 81 ea 86 f7
8b 04 90 11 95 1e 9a 2b 9f 38 a4 45 a9 52 ae 5f
b3 6c b8 79 bd 86 c2 93 c7 a0 cc ad d1 ba d6 c7
db d4 e0 e1 e5 ee
EOF
$PART_HEADER = <<EOF;
00 00 0F 27 1E 4E 2D 75  3C 9C 4B C3 5A EA 69 11
78 38 87 5F 96 86 A5 AD  B4 D4 C3 FB D2 22 E1 49
F0 70 FF 97 0E BF 1D E6  2C 0D 3B 34 4A 5B 59 82
68 A9 77 D0 86 F7 95 1E  A4 45 B3 6C C2 93 D1 BA
E0 E1 EF 08 FE 2F 0D 57  1C 7E 2B A5 3A CC 49 F3
58 1A 67 41 76 68 85 8F  94 B6 A3 DD B2 04 C1 2B
D0 52 DF 79 EE A0 FD C7  0C EF 1B 16 2A 3D 39 64
48 8B 57 B2 66 D9 75 00  84 27 93 4E A2 75 B1 9C
C0 C3 CF EA DE 11 ED 38  FC 5F 0B 87 1A AE 29 D5
38 FC 47 23 56 4A 65 71  74 98 83 BF 92 E6 A1 0D
B0 34 BF 5B CE 82 DD A9  EC D0 FB F7 0A 1F 19 46
28 6D 37 94 46 BB 55 E2  64 09 73 30 82 57 91 7E
A0 A5 AF CC BE F3 CD 1A  DC 41 EB 68 FA 8F 09 B7
18 DE 27 05 36 2C 45 53  54 7A 63 A1 72 C8 81 EF
90 16 9F 3D AE 64 BD 8B  CC B2 DB D9 EA 00 F9 27
08 4F 17 76 26 9D 35 C4  44 EB 53 12 62 39 71 60
EOF
$PART_HEADER_OFFSET = 0x4400;
$PART_TABLE_OFFSET = 0x4500;
$PART_END_MARK = 0x8000;
$VERSION = '1.4';
$DEBUG = 0;

sub new {
    my($class, $file, $part, $do_write) = @_;

    my $this = { file => $file, part => $part };
    bless $this;

    # open the AKAI disk with read-write enabled
    $this->{fh} = gensym;
    open($this->{fh}, ($do_write ? "+<$file" : $file)) || return;
    binmode($this->{fh});

    # automatically flush after print
    $_ = select($this->{fh});
    $| = 1;
    select($_);
    
    # move to the PART-th partition
    $this->{offset} = 0;
    while (1) {
	$this->seek(0);
	$_ = '';
	$this->read(\$_, 2);
	$this->{size} = unpack('v', $_);
	($! = ENODEV, return) if ($this->{size} == $PART_END_MARK);
	last if ($part <= 1);
	# add offset to point to the start of the partition
	$this->{offset} += $this->{size} * $BLOCK_SIZE;
	$part--;
    }

    # check the available space
    $this->{capacity} = $this->{size}; # should be rewritten

    $this;
}

sub DESTROY {
    my $this = shift;

    close($this->{fh});
}

sub seek {
    my($this, $pos) = @_;
    print STDERR "  seek($this, $pos)...\n" if $DEBUG > 2;

    seek($this->{fh}, $this->{offset} + $pos, 0) || return;
}

sub read {
    my($this, $varref, $size) = @_;
    print STDERR "  read($this, $varref, $size)...\n" if $DEBUG > 2;

    read($this->{fh}, $$varref, $size);
    1;
}

sub write {
    my($this, $var) = @_;
    print STDERR "  write($this, \$var)...\n" if $DEBUG > 2;

    print {$this->{fh}} $var;
    1;
}

# Convert ASCII string to AKAI encoded string.  All lowercase
# characters are changed to uppercase. 
sub ascii_to_akai {
    $_ = shift;
    
    $_ = sprintf '%-12s', uc $_;
    tr/0-9 A-Z\#+\-./\x00-\x28/;
    tr/\x29-\xff/ /;
    $_;
}

# Convert AKAI encoded string into ASCII string.  After decoding, all
# trailing spaces are removed.
sub akai_to_ascii {
    $_ = shift;

    tr/\x00-\x28/0-9 A-Z\#+\-./;
    s/\s+$//;
    $_;
}

# Make all lowercase characters to uppercase, and replace all
# underscores with spaces.  All invalid characters are replaced by
# spaces. For conformity, add a prceeding slash if the file name does
# not have. 
sub qualify_path {
    $_ = shift;

    s/\s+$//;
    tr/a-z_/A-Z /;
    $_ = '/' . $_ unless (m:^/:);
    $_;
}    

sub unqualify_path {
    $_ = shift;

    s/\s+$//;
    tr/A-Z /a-z_/;
    $_;
}

sub join_path {
    my($path1, $path2) = @_;

    $path1 =~ s:/+$::;
    $path2 =~ s:^/+::;
    "$path1/$path2";
}


# Get the POS-th entry in the file/directory entry at BLOCK.  The
# first block is considered as the directory entry starting from
# $DIR_ENTRY_OFFSET. 
sub read_entry {
    my($this, $block, $pos) = @_;
    print STDERR " read_entry($this, $block, $pos)...\n" if $DEBUG > 1;

    # return cached entry if it exists.
    if (exists $this->{entry_cache}->{$block}->{$pos}) {
	return %{$this->{entry_cache}->{$block}->{$pos}};
    }

    my($name, $type, $start, $size);
    if ($block == $ROOT_ENTRY_OFFSET) {
	$this->seek($DIR_ENTRY_OFFSET + $pos * $DIR_ENTRY_SIZE);
	$_ = '';
	$this->read(\$_, $DIR_ENTRY_SIZE);
	($name, $type, $start) = unpack('a12vv', $_);
	$size = 0;
    } else {
	if ($pos < 341) {
	    $this->seek($block * $BLOCK_SIZE + $pos * $FILE_ENTRY_SIZE);
	} else {
	    $_ = $this->read_fat($block);
	    $this->seek($_ * $BLOCK_SIZE + ($pos - 341) * $FILE_ENTRY_SIZE);
	}
	$_ = '';
	$this->read(\$_, $FILE_ENTRY_SIZE);
	($name, $type, @_) = unpack('a12x4C6', $_);
	$size  = $_[0] + ($_[1] << 8) + ($_[2] << 16);
	$start = $_[3] + ($_[4] << 8);
    }
    $name = akai_to_ascii($name);

    my %entry;
    if ($type) {
	print STDERR " read_entry: found $name, $type, $start, $size\n"
	    if $DEBUG > 1;
	%entry = (name => $name, type => $type, start => $start, 
		  size => $size, where => $block, at => $pos);
    }
    $this->{entry_cache}->{$block}->{$pos} = { %entry };

    return %entry;
}

# Real all entries in the file/directory entry located at BLOCLK.
# Those entries are returned as an array of hashes.
sub read_all_entries {
    my($this, $block) = @_;
    print STDERR "read_all_entries($this, $block)...\n" if $DEBUG;

    my %entries;
    my $nentries;
    if ($block == $ROOT_ENTRY_OFFSET) {
	$nentries = $MAX_DIR_ENTRIES;
    } else {
	$_ = $this->read_fat($block);
	$nentries = ($_ == $FAT_MARK_RESERVED) ?
	    $MAX_FILE_ENTRIES_S1000 : $MAX_FILE_ENTRIES_S3000;
    }
    for (0 .. $nentries - 1) {
	my %entry = $this->read_entry($block, $_);
	if (exists $entry{type}) {
	    $entries{$entry{name}} = { %entry };
	}
    }
    %entries;
}

# Find a free entry in the file/directory entry located at BLOCK, and
# return the position of it.  If not found, undef is returned. 
sub find_free_entry {
    my($this, $block) = @_;
    print STDERR " find_free_entry($this, $block)...\n" if $DEBUG > 1;

    my $nentries;
    if ($block == $ROOT_ENTRY_OFFSET) {
	$nentries = $MAX_DIR_ENTRIES;
    } else {
	$_ = $this->read_fat($block);
	$nentries = ($_ == $FAT_MARK_RESERVED) ?
	    $MAX_FILE_ENTRIES_S1000 : $MAX_FILE_ENTRIES_S3000;
    }
    my $i;
    for $i (0 .. $nentries - 1) {
	my %entry = $this->read_entry($block, $i);
	return $i unless (exists $entry{type});
    }
    undef;
}

# Overwrite the POS-th file/directory entry at BLOCK with the contents
# in the hash ENTRY.  Be careful since this code does NOT check
# whether the POS-th entry is already used or not. 
sub write_entry {
    my($this, $block, $pos, %entry) = @_;
    print STDERR " write_entry($this, $block, $pos, %entry)...\n" 
	if $DEBUG > 1;
    confess unless (defined($block) and defined($pos));

    # delete chaced entry
    delete $this->{entry_cache}->{$block}->{$pos};

    $entry{name} = ascii_to_akai($entry{name}); # if $entry{name};
    if ($block == $ROOT_ENTRY_OFFSET) {
	$this->seek($DIR_ENTRY_OFFSET + $pos * $DIR_ENTRY_SIZE);
	$_ = pack('A12vv', $entry{name}, $entry{type}, $entry{start});
    } else {
	if ($pos < 341) {
	    $this->seek($block * $BLOCK_SIZE + $pos * $FILE_ENTRY_SIZE);
	} else {
	    $_ = $this->read_fat($block);
	    $this->seek($_ * $BLOCK_SIZE + ($pos - 341) * $FILE_ENTRY_SIZE);
	}
	$_ = pack('A12x4CVXv@24', $entry{name}, $entry{type}, 
		  $entry{size}, $entry{start}); 
    }
    $this->write($_);
    1;
}

sub delete_entry {
    my($this, $block, $pos) = @_;
    print STDERR " delete_entry($this, $block, $pos)...\n" if $DEBUG > 1;

    my %entry = (name => '', type => $TYPE_FREE, start => 0, size => 0); 
    $this->write_entry($block, $pos, %entry);
}


sub read_fat {
    my($this, $block) = @_;
    print STDERR "  read_fat($this, $block)...\n" if $DEBUG > 2;

    # return cached entry if it exists
    if (exists $this->{fat_cache}->{$block}) {
	return $this->{fat_cache}->{$block};
    }

    $this->seek($FAT_OFFSET + $block * 2);
    $_ = '';
    $this->read(\$_, 2);
    my $val = unpack('v', $_);
    $this->{fat_cache}->{$block} = $val;

    return $val;
}

# Search the first unused block from the block FROM, and return the
# found block number.
sub find_free_block {
    my($this, $from) = @_;
    print STDERR "  find_free_block($from)...\n" if $DEBUG > 2;

    my $i;
    for $i ($from .. $this->{size} - 1) {
	return $i if ($this->read_fat($i) == $FAT_MARK_FREE);
    }
    $! = ENOSPC;
    undef;
}

sub write_fat {
    my($this, $block, $val) = @_;
    print STDERR "  write_fat($this, $block, $val)...\n" if $DEBUG > 2;
    confess unless (defined($block) and defined($val));

    # delete chaced data
    delete $this->{fat_cache}->{$block};

    $this->seek($FAT_OFFSET + $block * 2);
    $this->write(pack('v', $val));
    1;
}

sub delete_fat {
    my($this, $block) = @_;
    print STDERR "  delete_fat($this, $block)...\n" if $DEBUG > 2;

    my $entry = $this->read_fat($block);
    $this->write_fat($block, $FAT_MARK_FREE) if ($entry != $FAT_MARK_FREE);
    $entry;
}


sub read_block {
    my($this, $block) = @_;
    print STDERR "  read_block($this, $block)...\n" if $DEBUG > 2;

    $this->seek($block * $BLOCK_SIZE);
    $_ = '';
    $this->read(\$_, $BLOCK_SIZE);
    $_;
}

sub write_block {
    my($this, $block, $buf) = @_;
    print STDERR "  write_block($this, $block, buf)...\n" if $DEBUG > 2;
    confess unless (defined($block) and defined($buf));

    $this->seek($block * $BLOCK_SIZE);
    $this->write(substr($buf . ("\x00" x $BLOCK_SIZE), 0, $BLOCK_SIZE));
    1;
}


sub readdir {
    my($this, $dir) = @_;
    $dir = qualify_path($dir);
    print STDERR "readdir($this, $dir)...\n" if $DEBUG;

    my %entries;
    if ($dir eq '/') {
	%entries = $this->read_all_entries($ROOT_ENTRY_OFFSET);
    } else {
	my $parent = dirname($dir);
	my %parent = $this->readdir($parent);
	my $base = basename($dir);
	($! = ENOENT, return) unless (exists $parent{$base});
	%entries = $this->read_all_entries($parent{$base}->{start});
    }
    %entries;
}

# Obtain the property of the file spacified by PATH.  
sub stat {
    my($this, $path) = @_;
    $path = qualify_path($path);
    print STDERR "stat($this, $path)...\n" if $DEBUG;

    if ($path eq '/') {    
	return (name => '/', type => $TYPE_DIR_S3000,
		start => $ROOT_ENTRY_OFFSET);
    }
	
    my $dir = dirname($path);
    my %entries = $this->readdir($dir);
    my $base = basename($path);
    if (exists $entries{$base}) {
	return %{$entries{$base}};
    } else {
	$! = ENOENT;
	return ();
    }
}

# Return the property of the disk.  Currenty, only the available
# capacity is supported.
sub statfs {
    my $this = shift;
    print STDERR "stat_fs($this)...\n" if $DEBUG;

    return (file => $this->{file}, size => $this->{size}, 
	    capacity => $this->{capacity});
}


# Read the file at PATH from the disk, and write (append) it to the
# file specified by FILE.
sub read_file {
    my($this, $path, $file) = @_;
    $path = qualify_path($path);
    print STDERR "read($this, $path, $file)...\n" if $DEBUG;

    my %stat = $this->stat($path);
    ($! = ENOENT, return undef) unless $stat{type};

    # open oputput file
    my $fh = gensym;
    open($fh, ">>$file") || return;
    binmode($fh);
    
    my $block = $stat{start};
    my $size  = $stat{size};
    my $buf;
    while ($block != $FAT_MARK_END) {
	$buf = $this->read_block($block);
	if ($size < $BLOCK_SIZE) {
	    $buf = substr($buf, 0, $size);
	}
	$size -= $BLOCK_SIZE;
	print $fh $buf;
	$block = $this->read_fat($block);
    }

    close($fh);
    1;
}

# Open the file specified by FILE, and write it in the disk at PATH
# with TYPE.
sub write_file {
    my($this, $name, $path, $type) = @_;
    $path = qualify_path($path);
    print STDERR "write($this, $name, $path, $type)...\n" if $DEBUG;
    confess unless (defined($name) and defined($path) and defined($type));

    # make sure PATH does not exist
    my %stat = $this->stat($path);
    ($! = EEXIST, return) if (exists $stat{type});
    
    # open the input file 
    my $fh = gensym;
    open($fh, $name) || return;
    binmode($fh);		# for MS-DOS compatibility
    seek($fh, 0, 2);
    my $size = tell($fh);

    # the underlying directory must exist
    my $dir     = dirname($path);
    my %dirstat = $this->stat($dir);
    ($! = ENOENT,  return) if (!exists $dirstat{type});
    ($! = EBADF,   return) if ($type == $TYPE_FREE);
    ($! = ENOTDIR, return) unless ($dirstat{type} == $TYPE_DIR_S3000);

    # find a free entry in the directory, and write a new entry there
    my $start = $this->find_free_block(0) || return;
    my $pos   = $this->find_free_entry($dirstat{start});
    my %entry = (name  => basename($path),
		 type  => $type,
		 start => $start,
		 size  => $size);

    # before writing, check the disk capacity
    my $nblock = int($size / $BLOCK_SIZE) + 1;
    my $n = $start;
    for (0 .. $nblock - 1) {
	$n = $this->find_free_block($n + 1);
	($! = ENOSPC, return) unless $n;
    }

    $this->write_entry($dirstat{start}, $pos, %entry);

    # write the contents of the input file from $block, and update
    # corresponding FAT entry
    my $block = $entry{start};
    my $buf = '';
    my $next;
    seek($fh, 0, 0);		# move to the top
    while (1) {
	read($fh, $buf, $BLOCK_SIZE);
	$size -= $BLOCK_SIZE;
	$this->write_block($block, $buf);

	last if ($size <= 0);
	$next = $this->find_free_block($block + 1);
	$this->write_fat($block, $next);
	$block = $next;
    }
    $this->write_fat($block, $FAT_MARK_END);

    # close input file
    close($fh);
    1;
}

# Delete the file at PATH from the disk.
sub unlink {
    my($this, $path) = @_;
    $path = qualify_path($path);
    print STDERR "unlink($this, $path)...\n" if $DEBUG;

    # the file must exist
    my %stat = $this->stat($path);
    ($! = ENOENT, return undef) unless (exists $stat{type});

    # actually, delete the corresponding file entry
    $this->delete_entry($stat{where}, $stat{at});

    # clear all FAT chain starting from $stat{from}.
    my $block = $stat{start};
    my $nblocks = int($stat{size} / $BLOCK_SIZE) + 1;
    my $n;
    for $n (0 .. $nblocks - 1) {
	$block = $this->delete_fat($block);
    }
    confess unless ($block != $FAT_MARK_END);
    1;
}

sub mkdir {
    my($this, $dir) = @_;
    $dir = qualify_path($dir);
    print STDERR "mkdir($this, $dir)...\n" if $DEBUG;
    
    # check if the directry already exists
    my %stat = $this->stat($dir);
    ($! = EEXIST, return undef) if (exists $stat{type});

    # allocate a new directory entry
    my $block1 = $this->find_free_block(0) || return;
    my $block2 = $this->find_free_block($block1 + 1) || return;
    $this->write_fat($block1, $block2);
    $this->write_fat($block2, $FAT_MARK_RESERVED2);

    # initialize all file entries
    $this->write_block($block1, "\x00" x $BLOCK_SIZE);
    $this->write_block($block2, pack('x4048 C10 x4134',	# tune/level params
				     0x00, 0x01, 0x01, 0x00, 0x00,
				     0x00, 0x32, 0x09, 0x0c, 0xff));
    # write the directory entry in the parent directory
    my $parent  = dirname($dir);
    my %parent  = $this->stat($parent);
    my $pos     = $this->find_free_entry($parent{start});
    my $base    = basename($dir);
    my %entry   = (name => $base, type => $TYPE_DIR_S3000,
		   start => $block1);
    $this->write_entry($parent{start}, $pos, %entry);
    1;
}

sub format {
    my($this, @sizes) = @_;
    print STDERR "format($this, @sizes)...\n" if $DEBUG;

    # force to read/write from beginning
    $this->{offset} = 0;

    my $size;
    for $size (@sizes) {
	# each partition must be smaller than $MAX_PART_SIZE
	($! = EFBIG, return) if ($size > $MAX_PART_SIZE);

	# compose disk header
	my $head = pack('v', $size * 0x80);
	for (split(/[\s\n]+/, $DISK_HEADER)) {
	    $head .= pack('H2', $_);
	}
	$head .= pack('CCCC', ($size % 2) ? 0x55 : 0xd5,
		      0xba + (($size - 1) >> 1), 0x2f, 0x00);

	# write disk header
	$this->seek(0);
	$this->write($head);

	# compose partition header
	$head = '';
	for (split(/[\s\n]+/, $PART_HEADER)) {
	    $head .= pack('H2', $_);
	}

	# write partition header
	$this->seek($PART_HEADER_OFFSET);
	$this->write($head);
	
	# compose partition table
	$head = pack('C2', ($#sizes + 1), 0x01);
	my $total = 0;
	for (@sizes) {
	    $head .= pack('v', $_ * 0x80);
	    $total += $_ * 0x80;
	}
	$head .= pack('v', $total);
	$head .= "\x00" x (0x28 - length($head)); # padding zeros
	$_ = 0x66 - ($total / 0x80) * 4; # check sum??
	$head .= pack('C4', $_, 0x03, $_, 0x03);

	# write partition table
	$this->seek($PART_TABLE_OFFSET);
	$this->write($head);

	# initialize directory entries
	for (0 .. $MAX_DIR_ENTRIES - 1) {
	    $this->delete_entry($ROOT_ENTRY_OFFSET, $_);
	}

	# initializze file allocation table
	for (0 .. 2) {
	    $this->write_fat($_, $FAT_MARK_RESERVED);
	}
	for (3 .. ($size * 0x80 - 1)) {
	    $this->delete_fat($_);
	}

	# change offset to the next partition
	$this->{offset} += $size * 0x80 * $BLOCK_SIZE;

	# eliminate cached data
	delete $this->{entry_cache};
	delete $this->{fat_cache};
    }

    # write partition end mark
    $this->seek(0);
    $this->write(pack('v@2048', $PART_END_MARK));

    # initialize variables for later use
    $this->{offset} = 0;
    $this->{size} = $sizes[0] * 0x80;
    $this->{capacity} = $this->{size};
    1;
}

1;
