#!/usr/local/bin/perl
# 
# Object class for manipulating Ensoniq EPS disk image.
# Copyright (c) 1997, Hiroyuki Ohsaki.
# All rights reserved.
# 
# $Id: EpsDisk.pm,v 1.3 1998/07/14 07:01:35 oosaki Exp $
# 

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

package Synth::EpsDisk;

require 5.000;
use File::Basename;
use Symbol;
use Carp;
use POSIX;
use strict;

use vars qw(@ISA @EXPORT @EXPORT_OK
	    $BLOCK_SIZE
	    $MAX_FILE_ENTRIES
	    $FILE_ENTRY_SIZE
	    $FAT_OFFSET
	    $FAT_MARK_END
	    $ROOT_ENTRY_OFFSET
	    $TYPE_DIR
	    $VERSION
	    $DEBUG);

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(qualify_path unqualify_path join_path);

$BLOCK_SIZE = 512;
$MAX_FILE_ENTRIES = 39;
$FILE_ENTRY_SIZE = 26;
$FAT_OFFSET = 0xa00;
$FAT_MARK_END = 1;
$ROOT_ENTRY_OFFSET = 3;
$TYPE_DIR = 2;
$VERSION = '1.0';
$DEBUG = 0;

sub new {
    my($class, $file) = @_;

    my $this = { file => $file };
    bless $this;
    
    # open the EPS disk with read only
    $this->{fh} = gensym;
    open($this->{fh}, "<$file") || return;
    binmode($this->{fh});

    # automatically flush after print
    $_ = select($this->{fh});
    $| = 1;
    select($_);
    
    # check if the disk image is in GKH format.  if yes, remember the
    # size of GKH header.
    seek($this->{fh}, 0, 0);
    my $buf = '';
    $this->read(\$buf, $BLOCK_SIZE);
    if (substr($buf, 0, 4) eq 'TDDF') {
	# Intel format (TDDFI) is assumed.  Change the following 'v'
	# with 'n' for TDDFM.
	my $ntags = unpack('v', substr($buf, 6, 2));
	my $offset = 8 + $ntags * 10;
	for (0 .. $ntags - 1) {
	    @_ = unpack('CCvv', substr($buf, 8 + $_ * 10, 10));
	    $offset += $_[2] if ($_[0] == 20 or $_[0] == 21);
	}
	$this->{offset} = $offset;
    } else {
	$this->{offset} = 0;
    }

    $this;
}

sub DESTROY {
    my $this = shift;

    close($this->{fh});
}

sub seek {
    my($this, $pos) = @_;
    print STDERR "  seek($this, $pos)...\n" if $DEBUG > 2;

    seek($this->{fh}, $this->{offset} + $pos, 0) || return;
}

sub read {
    my($this, $varref, $size) = @_;
    print STDERR "  read($this, $varref, $size)...\n" if $DEBUG > 2;

    read($this->{fh}, $$varref, $size);
    1;
}

# Make all lowercase characters to uppercase, and replace all
# underscores with spaces.  All invalid characters are replaced by
# spaces. For conformity, add a prceeding slash if the file name does
# not have. 
sub qualify_path {
    $_ = shift;

    s/\s+$//;
    tr/a-z_/A-Z /;
    $_ = '/' . $_ unless (m:^/:);
    $_;
}    

sub unqualify_path {
    $_ = shift;

    s/\s+$//;
    tr/A-Z /a-z_/;
    $_;
}

sub join_path {
    my($path1, $path2) = @_;

    $path1 =~ s:/+$::;
    $path2 =~ s:^/+::;
    "$path1/$path2";
}


# Get the POS-th entry in the file/directory entry at BLOCK.  The
# first block is considered as the directory entry starting from
# $DIR_ENTRY_OFFSET. 
sub read_entry {
    my($this, $block, $pos) = @_;
    print STDERR " read_entry($this, $block, $pos)...\n" if $DEBUG > 1;

    # return cached entry if it exists.
    if (exists $this->{entry_cache}->{$block}->{$pos}) {
	return %{$this->{entry_cache}->{$block}->{$pos}};
    }

    my ($name, $type, $start, $size, $contig);
    $this->seek($block * $BLOCK_SIZE + $pos * $FILE_ENTRY_SIZE);
    $_ = '';
    $this->read(\$_, $FILE_ENTRY_SIZE);
    ($type, $name, $size, $contig, $start) = unpack('x Ca12nnN', $_);
    $name =~ s/\s+$//;

    my %entry;
    if ($type) {
	print STDERR " read_entry: found $name, $type, $start, $size\n"
	    if $DEBUG > 1;
	%entry = (name => $name, type => $type, start => $start,
		  size => $size * $BLOCK_SIZE, where => $block, at => $pos);
    }
    $this->{entry_cache}->{$block}->{$pos} = { %entry };

    return %entry;
}

# Real all entries in the file/directory entry located at BLOCLK.
# Those entries are returned as an array of hashes.
sub read_all_entries {
    my($this, $block) = @_;
    print STDERR "read_all_entries($this, $block)...\n" if $DEBUG;

    my %entries;
    my %entry;
    for (0 .. $MAX_FILE_ENTRIES - 1) {
	my %entry = $this->read_entry($block, $_);
	if (exists $entry{type}) {
	    $entries{$entry{name}} = { %entry };
	}
    }
    %entries;
}


sub read_fat {
    my($this, $block) = @_;
    print STDERR "  read_fat($this, $block)...\n" if $DEBUG > 2;

    # return cached entry if it exists
    if (exists $this->{fat_cache}->{$block}) {
	return $this->{fat_cache}->{$block};
    }

    $this->seek($FAT_OFFSET + int($block / 170) * $BLOCK_SIZE
		+ ($block % 170) * 3);
    $_ = '';
    $this->read(\$_, 3);
    my $val = unpack('N', "\x00$_");
    $this->{fat_cache}->{$block} = $val;

    return $val;
}


sub read_block {
    my($this, $block) = @_;
    print STDERR "  read_block($this, $block)...\n" if $DEBUG > 2;

    $this->seek($block * $BLOCK_SIZE);
    $_ = '';
    $this->read(\$_, $BLOCK_SIZE);
    $_;
}


sub readdir {
    my($this, $dir) = @_;
    $dir = qualify_path($dir);
    print STDERR "readdir($this, $dir)...\n" if $DEBUG;

    my %entries;
    if ($dir eq '/') {
	%entries = $this->read_all_entries($ROOT_ENTRY_OFFSET);
    } else {
	my $parent = dirname($dir);
	my %parent = $this->readdir($parent);
	my $base = basename($dir);
	($! = ENOENT, return) unless (exists $parent{$base});
	%entries = $this->read_all_entries($parent{$base}->{start});
    }
    %entries;
}

# Obtain the property of the file spacified by PATH.  
sub stat {
    my($this, $path) = @_;
    $path = qualify_path($path);
    print STDERR "stat($this, $path)...\n" if $DEBUG;

    if ($path eq '/') {
	return (name => '/', type => $TYPE_DIR,
		start => $ROOT_ENTRY_OFFSET);
    }
	
    my $dir = dirname($path);
    my %entries = $this->readdir($dir);
    my $base = basename($path);
    if (exists $entries{$base}) {
	return %{$entries{$base}};
    } else {
	$! = ENOENT;
	return ();
    }
}

# Read the file at PATH from the disk, and write (append) it to the
# file specified by FILE.
sub read_file {
    my($this, $path, $file) = @_;
    $path = qualify_path($path);
    print STDERR "read($this, $path, $file)...\n" if $DEBUG;

    my %stat = $this->stat($path);
    ($! = ENOENT, return undef) unless $stat{type};

    # open oputput file
    my $fh = gensym;
    open($fh, ">>$file") || return;
    binmode($fh);
    
    my $block = $stat{start};
    my $size  = $stat{size};
    my $buf;
    while ($block != $FAT_MARK_END) {
	$buf = $this->read_block($block);
	if ($size < $BLOCK_SIZE) {
	    $buf = substr($buf, 0, $size);
	}
	$size -= $BLOCK_SIZE;
	print $fh $buf;
	$block = $this->read_fat($block);
    }

    close($fh);
    1;
}

1;
