#!/usr/local/bin/perl
# 
# Object class for manipulating Ensoniq EPS instrument file.
# Copyright (c) 1997, Hiroyuki Ohsaki.
# All rights reserved.
# 
# $Id: EpsSample.pm,v 1.2 1998/07/03 14:32:43 oosaki Exp $
# 

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

package Synth::EpsSample;

require 5.000;
use Symbol;
use strict;

use vars qw(@ISA @EXPORT
	    $COMMON_DATA_SIZE
	    $LAYER_DATA_SIZE
	    $LAYER_TABLE_OFFSET
	    $WAVE_HEADER_SIZE
	    $WAVE_TABLE_OFFSET);

require Exporter;
@ISA = qw(Exporter);

$COMMON_DATA_SIZE = 655;
$LAYER_DATA_SIZE = 224;
$LAYER_TABLE_OFFSET = 0x64;
$WAVE_HEADER_SIZE = 288;
$WAVE_TABLE_OFFSET = 0x84;

sub new {
    my $class = shift;

    my $this = {};
    bless $this;
    $this;
}

sub eps_to_ascii {
    my $old = shift;

    my $new = '';
    for (0 .. 11) {
	$new .= substr($old, $_ << 1, 1);
    }
    $new =~ s/\s+$//;
    $new;
}

sub seek {
    my($this, $pos, $whence) = @_;

    $whence = 0 unless defined($whence);
    seek($this->{fh}, $this->{offset} + $pos, $whence) || return;
}

sub load {
    my($this, $file) = @_;

    # open instrument file
    $this->{fh} = gensym;
    open($this->{fh}, $file) || return;
    binmode($this->{fh});
    
    # skip INS header if it exists
    $_ = '';
    seek($this->{fh}, 2, 0);
    read($this->{fh}, $_, 9);
    $this->{offset} = ($_ eq 'Eps File:') ? 512 : 0;
}

sub read_common {
    my $this = shift;

    $this->seek(0);
    $_ = '';
    read($this->{fh}, $_, $COMMON_DATA_SIZE);
    @_ = unpack('nn nNa24 vvv nv vvvv vv vv v vvv', $_);
    my $name = eps_to_ascii($_[4]);

    (name         => $name,
     midi_prog    => $_[6],
     midi_chan    => $_[5],
     patch_select => [ @_[10 .. 13] ],
     low_key      => $_[19],
     high_key     => $_[20]);
}

sub read_layer_address {
    my($this, $no) = @_;

    $this->seek($LAYER_TABLE_OFFSET + ($no << 2));
    $_ = '';
    read($this->{fh}, $_, 4);
    @_ = unpack('CxCC', $_);
    ($_[2] << 16) + ($_[0] << 12) + ($_[1] << 4);
}

sub read_layer {
    my($this, $no) = @_;

    my $address = $this->read_layer_address($no);
    return () unless $address;

    $this->seek($address);
    my $buf = '';
    read($this->{fh}, $buf, $LAYER_DATA_SIZE);
    @_ = unpack('nnnv x2 a24 x6 vvvv', $buf);
    my $size = ($_[0] >> 4) + ($_[1] << 8);
    my $name = eps_to_ascii($_[4]);
    my @keymap = unpack('v*', substr($buf, 0x30, 88 * 2));

    (number   => $no,
     address  => $address,
     name     => $name,
     low_vel  => $_[5],
     high_vel => $_[6],
     pitch    => $_[7],
     delay    => $_[8],
     keymap   => [ @keymap ],
     );
}

sub dump_layer {
    my($this, %layer) = @_;

    print <<EOF;
number   => $layer{number},
address  => $layer{address},
name     => $layer{name},
low_vel  => $layer{low_vel},
high_vel => $layer{high_vel},
pitch    => $layer{pitch},
delay    => $layer{delay},
EOF
    print "keymap   =>";
    for $_ (0 .. 87) {
	print "\n" if ($_ % 12 == 0);
	printf "%3d ", $layer{keymap}[$_];
    }
    print "\n\n";
}

sub read_wave_address {
    my($this, $no) = @_;

    $this->seek($WAVE_TABLE_OFFSET + ($no << 2));
    $_ = '';
    read($this->{fh}, $_, 4);
    @_ = unpack('CxCC', $_);
    ($_[2] << 16) + ($_[0] << 12) + ($_[1] << 4);
}

sub read_wave_header {
    my($this, $no) = @_;
    
    my $address = $this->read_wave_address($no);
    return () unless $address;

    $this->seek($address);
    my $buf = '';
    read($this->{fh}, $buf, $WAVE_HEADER_SIZE);

    @_ = unpack('nnn x4 a24 v', $buf);
    my $size = (($_[0] >> 4) + ($_[1] << 8) - $WAVE_HEADER_SIZE) >> 1;
    my $name = eps_to_ascii($_[3]);
    my $copy = $_[4];

    my $key  = unpack('C', substr($buf, 0xaa, 1));
    my $tune = unpack('c', substr($buf, 0xb6, 2));
    my $vol  = unpack('C', substr($buf, 0xd0, 1));
    my $pan  = unpack('c', substr($buf, 0xdc, 1));
    my $pan_frac = unpack('c', substr($buf, 0xdd, 1));

    @_ = unpack('v vvvv vvvv vvvv vvvv v vv', 
		substr($buf, 0xee, 40)); 
    my $start      = (($_[1]  << 16) + ($_[2]  << 8) + $_[3] ) >> 1;
    my $end        = (($_[5]  << 16) + ($_[6]  << 8) + $_[7] ) >> 1;
    my $loop_start = (($_[9]  << 16) + ($_[10] << 8) + $_[11]) >> 1;
    my $loop_end   = (($_[13] << 16) + ($_[14] << 8) + $_[15]) >> 1;

    (number     => $no,
     address    => $address + $WAVE_HEADER_SIZE,
     size       => $size,
     name       => $name,
     copy       => $copy,
     rate       => $_[17] & 0xff,
     key        => $key,
     tune       => $tune,
     start      => $start,
     end        => $end,
     loop_mode  => $_[0],
     loop_start => $loop_start,
     loop_end   => $loop_end,
     vol        => $vol,
     pan        => $pan & 0xf,
     pan_frac   => $pan_frac,
     low_key    => $_[18],
     high_key   => $_[19],
     );
}

sub read_wave_data {
    my($this, $no) = @_;

    my %wave = $this->read_wave_header($no);
    return undef unless (exists $wave{name});

    # return if this wave is a copy of other sample
    return undef if ($wave{copy} != 0);

    $this->seek($wave{address});
    my $data = '';
    read($this->{fh}, $data, $wave{size} << 1);
    return \$data;
}

sub dump_wave {
    my($this, %wave) = @_;

    print <<EOF;
number     => $wave{number},
address    => $wave{address},
size       => $wave{size},
name       => $wave{name},
copy       => $wave{copy},
rate       => $wave{rate},
key        => $wave{key},
tune       => $wave{tune},
start      => $wave{start},
end        => $wave{end},
loop_mode  => $wave{loop_mode},
loop_start => $wave{loop_start},
loop_end   => $wave{loop_end},
vol        => $wave{vol},
pan        => $wave{pan},
pan_frac   => $wave{pan_frac},
low_key    => $wave{low_key},
high_key   => $wave{high_key},

EOF
}
